#!/usr/bin/python

"""//Cloud Worker implementation using command line execution//

- > Contains some function to start and stop Cloud Worker instance, using the command line execution from configuration file

- File: command_cloudworker.py

- Notes: 
- The Cloud configuration file associated to the managed instance must contain the command to start (CMD_CW_START) and to stop (CMD_CW_STOP) the Cloud worker. 

 
"""

import threading
import time

import sys
import os
sys.path.append(os.path.join(sys.path[0],".."))
os.chdir(sys.path[0])

from tools import conf_func
from tools import log_func
log=log_func.get_spequlos_log(__file__)


#Timeout for command execution to complete
CMD_EXEC_TIMEOUT=600



def read_cloud_params(cloud_config_file):
    cloud_params = conf_func.read_config(cloud_config_file)
    for p in ["CMD_CW_START", "CMD_CW_STOP" ]:
        if not (p in cloud_params):
            raise Exception("Missing parameters %s in %s" % (p,cloud_config_file))
        
    return cloud_params

def exec_cmd(cmd,env,print_output=False):
    #Adding 'export VAR=VAL' to command from content of env dictionnary
    for v in env:
        var=v.strip('\'" ')
        var=var.replace("'","\\\'")
        var=var.replace('"','\\\"')
        var=var.replace(' ','_')
        val=env[v].strip('\'" ')
        val=val.replace('"','\\\"')

        export_cmd='export %s="%s";' % (var,val)
        cmd=export_cmd+cmd 

    #Adding 'echo $SQS_DATA' at the end of the command to allow exporting information from command to SpeQuloS
    cmd=cmd+";echo '';echo $SQS_DATA"

    log.info("Executing command: %s" % cmd)    
    import subprocess
    p=subprocess.Popen(cmd,shell=True,stdout=subprocess.PIPE)
    
    waited_time=0
    ret=None
    while ret is None and waited_time<CMD_EXEC_TIMEOUT:
        #$SQS_DATA is the last thing printed on stdout
        for l in p.stdout:
            if print_output:
                print l
            sqs_data=l
        
        ret=p.poll()
        if ret is None:
            time.sleep(1)
            waited_time+=1

    p.stdout.close()

    if ret is None:
        log.error("Command execution is too long to complete")
        log.error("Sending KILL signal")
        p.kill()
        raise Exception("Command execution is too long to complete")

    elif ret!=0:
        log.warn("Command execution return code is not 0 (%s)" % str(ret))
        log.warn("Continuing anyway...")
        
    return (ret,sqs_data.replace("\n",""))
    



class Command_CW_Starter(threading.Thread):
    """//Thread that starts a Cloud worker using command line execution//

    - > The value of CMD_CW_START option in Cloud service configuration file will be executed 
    
    - Arguments:
     - cloud_id -- The Cloud service identifier
     - cloud_config_file -- The Cloud service configuration file
     - dg_id -- The supported Desktop Grid identifer
     - dg_config_file -- The Desktop Grid configuration file
     - batch_id -- The supported batch_id identifier
     - batch_data -- Data related to the supported batch. For XWHEP, it stores the batch UID.          
     - instance_id -- The identifier of the Cloud instance to start
     
    - Attributes:
     - execution_is_success -- When the thread is terminated, True if the the execution of the of the command succeed and the Cloud worker correctly started, False otherwise.
    
    """     
    
    def __init__(self,cloud_id,cloud_config_file,dg_id,dg_config_file,batch_id,batch_data,instance_id):
        threading.Thread.__init__(self)        
        if not isinstance(cloud_id,str):
            raise Exception("Invalid cloud_id argument (must be a string)")
        if not isinstance(cloud_config_file,str):
            raise Exception("Invalid cloud_config_file argument (must be a string)")
        if not isinstance(dg_id,str):
            raise Exception("Invalid dg_id argument (must be a string)")
        if not isinstance(dg_config_file,str):
            raise Exception("Invalid dg_config_file argument (must be a string)")
        if not isinstance(batch_id,str):
            raise Exception("Invalid batch_id argument (must be a string)")
      
        self.cloud_params = read_cloud_params(cloud_config_file)

        self.env_var={}
        self.env_var['SQS_CLOUD_ID']=str(cloud_id)
        self.env_var['SQS_CLOUD_CONF']=str(cloud_config_file)
        self.env_var['SQS_DG_ID']=str(dg_id)
        self.env_var['SQS_DG_CONF']=str(dg_config_file)
        self.env_var['SQS_BATCH_ID']=str(batch_id)
        self.env_var['SQS_BATCH_DATA']=str(batch_data)        
        self.env_var['SQS_INSTANCE_ID']=str(instance_id)
        
        self.execution_is_success=False
        self.specific_data=None


    def run(self):
        log.info("Start command is %s" % self.cloud_params["CMD_CW_START"])
        
        (ret,data)=exec_cmd(self.cloud_params["CMD_CW_START"],self.env_var)
        
        log.info("Command executed, return code is %s" % ret)
                        
        if ret==0:
            self.execution_is_success = True
            if data!="":
                self.specific_data=data



class Command_CW_Stopper(threading.Thread):
    """//Thread that stops a Cloud worker using command line execution//

    - > The value of CMD_CW_STOP option in Cloud service configuration file will be executed 
    
    - Arguments:
     - cloud_id -- The Cloud service identifier. Stored in $SQS_CLOUD_ID
     - cloud_config_file -- The Cloud service configuration file. Stored in $SQS_CLOUD_CONF
     - dg_id -- The supported Desktop Grid identifer. Stored in $SQS_DG_ID
     - dg_config_file -- The Desktop Grid configuration file. Stored in $SQS_DG_CONF
     - batch_id -- The supported batch_id identifier. Stored in $SQS_BATCH_ID
     - batch_data -- Data related to the supported batch. For XWHEP, it stores the batch UID.          
     - instance_id -- The identifier of the Cloud instance to start. Stored in $SQS_INSTANCE_ID
     - instance_data -- Arbitrary data. Stored in $SQS_DATA variable.
     
    - Attributes:
     - stop_is_success -- When the thread is terminated, True if the the execution of the of the command succeed and the Cloud worker correctly stopped, False otherwise.
    
    """     
    
    def __init__(self,cloud_id,cloud_config_file,dg_id,dg_config_file,batch_id,batch_data,instance_id,instance_data):
        threading.Thread.__init__(self)        
        if not isinstance(cloud_id,str):
            raise Exception("Invalid cloud_id argument (must be a string)")
        if not isinstance(cloud_config_file,str):
            raise Exception("Invalid cloud_config_file argument (must be a string)")
        if not isinstance(dg_id,str):
            raise Exception("Invalid dg_id argument (must be a string)")
        if not isinstance(dg_config_file,str):
            raise Exception("Invalid dg_config_file argument (must be a string)")
        if not isinstance(batch_id,str):
            raise Exception("Invalid batch_id argument (must be a string)")
      
        self.cloud_params = read_cloud_params(cloud_config_file)

        self.env_var={}
        self.env_var['SQS_CLOUD_ID']=str(cloud_id)
        self.env_var['SQS_CLOUD_CONF']=str(cloud_config_file)
        self.env_var['SQS_DG_ID']=str(dg_id)
        self.env_var['SQS_DG_CONF']=str(dg_config_file)
        self.env_var['SQS_BATCH_ID']=str(batch_id)
        self.env_var['SQS_BATCH_DATA']=str(batch_data)                
        self.env_var['SQS_INSTANCE_ID']=str(instance_id)
        self.env_var['SQS_DATA']=str(instance_data)
        
        self.stop_is_success=False


    def run(self):
        log.info("Stop command is %s" % self.cloud_params["CMD_CW_STOP"])
                
        (ret,data)=exec_cmd(self.cloud_params["CMD_CW_STOP"],self.env_var)
        
        log.info("Command executed, return code is %s" % ret)
        
        if ret==0:
            self.stop_is_success = True
        


if __name__=="__main__":
    import sys
    cw = Command_CW_Starter("cloud_none",sys.argv[1],"dg_none","/none/dg/config/file","batch_none","instance_none")
    cw.start()
    cw.join()
    if cw.execution_is_success:
        print "Press enter to destroy the cloud worker"
        x=raw_input()
        cwstop = Command_CW_Stopper("cloud_none",sys.argv[1],"dg_none","/none/dg/config/file","batch_none","instance_none",cw.specific_data)
        cwstop.start()
