#!/usr/bin/python

"""//Cloud Worker implementation using libclud//

- > Contains some function to start and stop Cloud Worker instance, using the libcloud API

- File: libcloud_cloudworker.py

- Notes: 
 - The Cloud configuration file associated to the managed instance must contain the libcloud driver to use (LC_DRIVER), optional Cloud server information (LC_HOST, LC_PORT, LC_PATH, LC_SECURE, LC_DUMMY_EC2), credentials (LC_KEY and LC_SECRET), instance information (CW_IMAGE_ID, CW_SIZE_ID) and SSH authentication information (CW_SSH_USER and CW_SSH_KEY or CW_SSH_PWD).
 
"""

import libcloud.compute.providers as LC_Providers
from libcloud.compute.drivers import ec2 as LC_DriverEC2
from libcloud.compute.types import NodeState as LC_NodeState
import paramiko

import time
import threading

import sys
import os
sys.path.append(os.path.join(sys.path[0],".."))
os.chdir(sys.path[0])

from tools import conf_func
from tools import log_func
log=log_func.get_spequlos_log(__file__)



#Passthrough for OpenNebula/OpenStack EC2 interfaces (taken from Nimbus driver)
def dummy_ex_describe_addresses(nodes):
    nodes_elastic_ip_mappings = {}
    for node in nodes:
        nodes_elastic_ip_mappings[node.id] = []
    return nodes_elastic_ip_mappings
def dummy_ex_create_tags(node, tags):
    pass
#Passthrough for a bug in EC2 driver related to authentication signature ONLY VALID FOR LIBCLOUD 0.5.2
def dummy_get_aws_auth_param(self, params, secret_key, path='/'):
    import urllib
    import base64
    import hmac
    from hashlib import sha256
    keys = params.keys()
    keys.sort()
    pairs = []
    for key in keys:
        pairs.append(urllib.quote(key, safe='') + '=' +
                     urllib.quote(params[key], safe='-_~'))

    qs = '&'.join(pairs)

    hostname = self.host
    portnum = str(self.port[int(self.secure)])
    if (self.secure and portnum != '443') or (not self.secure and portnum !='80'):
        hostname+=":"+portnum
    string_to_sign = '\n'.join(('GET', hostname, path, qs))
 
    b64_hmac = base64.b64encode(
        hmac.new(secret_key, string_to_sign, digestmod=sha256).digest()
    )
    return b64_hmac




def read_cloud_params(cloud_config_file):
    cloud_params = conf_func.read_config(cloud_config_file)
    for p in ["LC_DRIVER", "LC_KEY", "LC_SECRET", "CW_IMAGE_ID", "CW_SIZE_ID" ]:
        if not (p in cloud_params):
            raise Exception("Missing parameters %s in %s" % (p,cloud_config_file))

    if not ("CW_SSH_KEY" in cloud_params or "CW_SSH_PWD" in cloud_params):
        raise Exception("Missing parameters CW_SSH_KEY or CW_SSH_PWD in %s" % (p,cloud_config_file))

    if not "CW_SSH_USER" in cloud_params:
        cloud_params["CW_SSH_USER"] = "root" 
    
    if "LC_SECURE" in cloud_params:
        if cloud_params["LC_SECURE"] in ("No","0","False"):
            cloud_params["LC_SECURE"] = False
        else:
            cloud_params["LC_SECURE"] = True
   
    if "LC_DUMMY_EC2" in cloud_params:
        if cloud_params["LC_DUMMY_EC2"] in ("True","Yes","1"):
            cloud_params["LC_DUMMY_EC2"] = True
        else:
            cloud_params["LC_DUMMY_EC2"] = False
        
    return cloud_params


def LC_connect(cloud_params):
    log.info("Selecting libcloud driver for %s" % cloud_params["LC_DRIVER"])
    Driver = None
    for driver_num in LC_Providers.DRIVERS:
        if cloud_params["LC_DRIVER"] == LC_Providers.DRIVERS[driver_num][1]:
            Driver = LC_Providers.get_driver(driver_num)
    
    if Driver is None:
        raise Exception("No Libcloud driver with name %s" % cloud_params["LC_DRIVER"])

    #This build the dictionnary containing Driver __init__ arguments
    #Each possible argument that is found in cloud service configuration file (in uppercase and prefixed by "LC_") is used.
    import inspect
    lc_connection_args={}
    for arg in inspect.getargspec(Driver.__init__).args:
        if "LC_"+arg.upper() in cloud_params:
            lc_connection_args[arg]=cloud_params["LC_"+arg.upper()]

    log.info("Connection to driver %s using option %s" % (cloud_params["LC_DRIVER"], lc_connection_args))
    conn = Driver(**lc_connection_args) 

    #Passtrough for uncomplete EC2 interfaces (OneNebula/OpenStack), as used for Nimbus
    if ("LC_DUMMY_EC2" in cloud_params) and cloud_params["LC_DUMMY_EC2"]:
        conn.ex_describe_addresses=dummy_ex_describe_addresses
        conn.ex_create_tags=dummy_ex_create_tags
        #Uncomment this for authentication signature bug in libcloud 0.5.2
        #conn.connectionCls._get_aws_auth_param=dummy_get_aws_auth_param

    return conn
    

class Libcloud_CW_Starter(threading.Thread):
    """//Thread that starts a Cloud worker using libcloud and make it executes a list of commands//

    - > In the current implmementation, it makes few minutes for an instance to start.
    
    - Arguments:
     - cloud_config_file -- The Cloud service configuration file
     - list_of_commands -- The list of commands to be executed by the cloud worker, as a list of stings
     - instance_id -- The identifier of the Cloud instance to start     
     
    - Attributes:
     - execution_is_success -- When the thread terminated, True if the cloud worker correctly started and the execution of the list of commands succeed, False otherwise
     - specific_data -- When the thread terminated and the execution succeed, may contain some specific data related to the batch. It contains the instance identifier when libcloud is used. 
    
    """     
    
    def __init__(self,cloud_config_file,list_of_commands,instance_id):
        threading.Thread.__init__(self)
        if not isinstance(list_of_commands,list) or len(list_of_commands)==0:
            raise Exception("No command to execute")
        for s in list_of_commands:
            if not isinstance(s,str):
                raise Exception("Invalid list of commands (must be a list of string)")
        
        self.cloud_params = read_cloud_params(cloud_config_file)

        self.list_of_commands = list_of_commands
        self.instance_id = instance_id
        self.execution_is_success=False
        self.specific_data=None


    def run(self):
        conn = LC_connect(self.cloud_params)

        log.info("Retrieving the list of images")
        try:
            images = conn.list_images() 
        except Exception as e:
            log.error("Error when retrieving the list of images !")
            log.error(e)
            raise Exception("Unable to start Cloud worker: Error when retrieving the list of images")
        
        cw_img = None
        for i in images:
            if self.cloud_params["CW_IMAGE_ID"] == i.id or self.cloud_params["CW_IMAGE_ID"] == i.name:
                cw_img=i        

        if cw_img == None:
            log.error("Cannot find CloudWorker image in Cloud server")
            raise Exception("Unable to start Cloud worker: Cannot find CloudWorker image")

        log.info("CloudWorker image found as %s" % str(cw_img))

        log.info("Reading through the list of instance types to find the type we want for the CW")
        sizes = conn.list_sizes() 
        cw_size = None
        for s in sizes:
            if self.cloud_params["CW_SIZE_ID"] == s.id or self.cloud_params["CW_SIZE_ID"] == s.name:
                cw_size=s
        
        if cw_size == None:
            log.error("Cannot find CloudWorker size in Cloud server")
            raise Exception("Unable to start Cloud worker: Cannot find CloudWorker size")

        log.info("CloudWorker instance type found as %s" % str(cw_size))
        
        log.info("Starting the instance")
        cw_node_launch = conn.create_node(name='lc_cw-%s' % str(self.instance_id), image=cw_img, size=cw_size) 
        
        #As the instance has no IP address yet, we must wait for the starting process to be completed
        #Periodacally read the list of existing instances until it find a running instance with the CW instance name
        log.info("Waiting for the instance startup to be completed...")
        cw_node_started = None
        waited_time = 0
        while cw_node_started==None and waited_time<600:
            time.sleep(5)
            waited_time += 5
            nodes = conn.list_nodes()
            for n in nodes:
                if n.state == LC_NodeState.RUNNING and (n.name == cw_node_launch.name or n.uuid == cw_node_launch.uuid):
                    cw_node_started = n
        
        if cw_node_started==None:
            log.error("Cannot retrieve the started instance of the CloudWorker")
            log.error("Destroying the instance previously created")
            cw_node_launch.destroy()
            raise Exception("Unable to start Cloud worker: Cannot retrieve the started instance")

        log.info("The instance is started as %s" % cw_node_started)

        ip = (len(cw_node_started.public_ip)>0 and cw_node_started.public_ip[0])\
          or (len(cw_node_started.private_ip)>0 and cw_node_started.private_ip[0]) \
          or (cw_node_started.extra.has_key('dns_name') and cw_node_started.extra['dns_name']) \
          or (cw_node_started.extra.has_key('private_dns') and cw_node_started.extra['private_dns'])
          
        if ip is None:
            log.error("Cannot retrieve the CloudWorker IP address")
            log.error("Destroying the instance previously created")
            cw_node_launch.destroy()
            raise Exception("Unable to start Cloud worker: Cannot retrieve the CloudWorker IP address")
  
        self.specific_data = cw_node_started.id

        log.info("Connecting with SSH to %s" % str(ip))
        ssh = paramiko.SSHClient()
        ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())
        waited_time = 0
        while waited_time<600:
            try:
                if "CW_SSH_KEY" in self.cloud_params:
                    ssh.connect(ip,username=self.cloud_params["CW_SSH_USER"],key_filename=self.cloud_params["CW_SSH_KEY"])
                    break
                elif "CW_SSH_PWD" in self.cloud_params:
                    ssh.connect(ip,username=self.cloud_params["CW_SSH_USER"],password=self.cloud_params["CW_SSH_PWD"])
                    break
                else:
                    log.error("Cannot connect to CloudWorker")
                    log.error("Destroying the instance previously created")
                    cw_node_launch.destroy()                
                    raise Exception("Unable to configure Cloud worker: No SSH key or password provided")                    
            except Exception as e:
                log.warn("Failing to connect")
                log.warn(e)                
                if waited_time<600:
                    log.warn("Trying again in 5 seconds")
                    time.sleep(5)
                    waited_time+=5
                else:
                    log.error("Too much tries, unable to connect")
                    log.error("Destroying the CloudWorker instance")
                    cw_node_started.destroy()
                    raise Exception("Unable to configure Cloud worker: Cannot connect using SSH")                    
        
        log.info("Executing commands")        
        for cmd in self.list_of_commands:
            log.info("Executing %s" % cmd)
            (stdin, stdout, stderr) =  ssh.exec_command(cmd)
            stdin.close()
            log.info("The output is \n%s\n%s" % ("".join(stdout.readlines()),"".join(stderr.readlines())) ) 
        
        log.info("All commands executed, closing connection")
        self.execution_is_success = True
        ssh.close()


class Libcloud_CW_Stopper(threading.Thread):
    """//Thread that stops a cloud worker using libcloud//

    - Arguments:
     - cloud_config_file -- The Cloud service configuration file
     - instance_id -- The intance identifier of the cloud worker in cloud server
     
    - Attributes:
     - stop_is_success -- When the thread terminated, True if the cloud worker correctly stopped, False otherwise

    """     

    def __init__(self,cloud_config_file,instance_id):
        threading.Thread.__init__(self)
        if not isinstance(instance_id,str):
            raise Exception("The instance_id must be a string")
                 
        self.cloud_params = read_cloud_params(cloud_config_file)

        self.instance_id = instance_id
        self.stop_is_success = False
        
    def run(self):
        conn = LC_connect(self.cloud_params)
        
        log.info("Retrieving the list of instance")
        nodes = conn.list_nodes() 
        cw_node = None
        for n in nodes:
            if self.instance_id == n.id:
                cw_node=n

        if cw_node == None:
            log.error("Cannot find CloudWorker instance in Cloud server")
            raise Exception("Unable to stop Cloud worker: Cannot find CloudWorker instance")         
                       
        log.info("CloudWorker instance found as %s" % str(cw_node))

        r = cw_node.destroy()

        if r != True:
            log.error("Error when terminating the instance")
            raise Exception("Unable to stop Cloud worker: Unknown instance termination")                     
        else:
            log.info("Instance destroyed")
            self.stop_is_success=True





if __name__=="__main__":
    import sys
    cw = Libcloud_CW_Starter(sys.argv[1],['uname -a','pwd','ls'],0)
    cw.start()
    cw.join()
    if cw.execution_is_success:
        print "Press enter to destroy the cloud worker"
        x=raw_input()
        cwstop = Libcloud_CW_Stopper(sys.argv[1],cw.specific_data)
        cwstop.start()
