#! /usr/bin/python

"""//Program that monitor QoS-enabled batches//

- > This program manages batches previously created by the start_qos function by allocating cloud resources to batch being funded by users. It should be started periodically to perform monitoring operations. 

- File: monitor_batches.py

- Notes: 
 - A database configuration file named "db.cfg" must be present in the directory. 
 - A file named "creditsystem.cfg" containing a line "address=<url>", with <url> the url to access to the Credit System web server, must be present in the directory. 
 - A file named "oracle.cfg" containing a line "address=<url>", with <url> the url to access to the Oracle web server, must be present in the directory. 

- Algorithm:
``` 
For each of batches, this function checks 
  if we are at 90% of the batch completion
   and if the oracle-predicted benefit of using cloud 
                                              is enough, 
   and if the batch order has more credits than the 
                    credit cost predicted by the oracle, 
   and if there is one cloud worker available, 
      starting it and assigning the batch to it.
``` 

"""

import sys
import os
sys.path.append(os.path.join(sys.path[0],".."))
os.chdir(sys.path[0])
from tools import conf_func
from tools import web_func
from tools import db_func
from tools import log_func
log=log_func.get_spequlos_log(__file__)

import cloudworker
import oracle_rpc
import creditsystem_rpc
import dg_rpc

CW_START_THRESHOLD = 0.9


def monitor_batches(db):
    """//Function implementing the batches monitoring//
      
    - Arguments:
     - db -- The scheduler's DB connection, as a SpeQuloS_DB object

    """        
    log.info("Retrieving QoSBatches list...")
    
    qos_batches = db.sql_select(("QoSBatch","DesktopGrid"),('batch_id','dg_id','batch_data',"dg_conf_file"))
    
    if len(qos_batches)==0:
        log.info("No QoSBatches to manage, exiting...")
    else:
        for b in qos_batches:
            log.info("Managing batch %s @ %s" % (b['batch_id'],b['dg_id']))

            log.debug("Checking how many cloud workers are working on this batch")
            cworkers = db.sql_select("CloudWorker",('cloud_worker_id','cloud_service_id','cloud_worker_data','batch_id'), "batch_id='%s'" % b['batch_id'])
            log.info("%d Cloud workers are working on batch %s" % (len(cworkers),b['batch_id']))
        
            log.info("Checking batch completion...")            
            completion = oracle_rpc.get_percent_completed(b['batch_id'])
            if completion==None:
                log.warn("No result from get_percent_completed, skipping this batch")
            elif completion<CW_START_THRESHOLD:
                log.info("The batch completion is not advanced enough to launch cloud workers")
            elif completion>=1:
                log.info("The batch completion is too advanced to launch cloud workers")
                #TODO : Test this part
                if len(cworkers)==0:
                    log.info("No cloud workers are working on this batch, deregistering it...")
                    
                    log.info("Updating the scheduler database")
                    db.sql_delete("QoSBatch","batch_id='%s'" % b['batch_id'])
                    
                    log.info("Unconfigure grid %s for QoS on batch %s" % (b['dg_id'],b['batch_id']))
                    dg_rpc.unconfigure_QoS(b['batch_id'],b['dg_id'],b['dg_conf_file'],b['batch_data'])
            else:
                log.info("The batch completion is higher than %s checking benefits of using cloud..." % CW_START_THRESHOLD)

                (benef,cost) = oracle_rpc.calculate_cloud_benefit(b['batch_id'])
                if benef==None:
                    log.error("No result from calculate_cloud_benefit, skipping this batch")
                elif benef<MIN_BENEFIT:
                    log.info("The benefit of using cloud is not sufficient, skipping this batch")
                else:
                    log.info("This batch can benefit from the use of cloud  ! Checking if there is enough credits ordered...")

                    cred = creditsystem_rpc.get_credits(b['batch_id'])
                    if cred==None:
                        log.error("No result from get_credit, skipping this batch")
                    elif cred<cost:
                        log.info("Not enough credits ordered to use cloud ressources, skipping this batch")
                    else:
                        log.info("There is enough credits to use cloud ressources !" +\
                                                "Checking cloud worker availability...")                            

                        available_cworkers = db.sql_select(("CloudService","CloudWorker"),('cloud_service_id','cloud_service_conf_file','handler','cloud_worker_id'),"batch_id IS NULL")
                        
                        if len(available_cworkers)==0:
                            log.warning("No cloud worker available!! Cannot do anything for this batch now.")
                        else:
                            cw = available_cworkers[0]    
                            log.info("Cloud worker %s@%s (%s) is available" % (cw['cloud_worker_id'], cw['cloud_service_id'], cw['handler']))
                            
                            log.info("Assigning it to %s" % (b['batch_id']))
                            db.sql_update("CloudWorker",('batch_id',),(b['batch_id'],),"cloud_worker_id='%s'" % cw['cloud_worker_id'])
                            
                            log.info("Starting it !")
                            result_data = cloudworker.start_cloud_worker(cw['cloud_service_id'], \
                                cw['cloud_service_conf_file'],b['dg_id'],b['dg_conf_file'],b['batch_id'],b['batch_data'],cw['cloud_worker_id'],cw['handler'])
                            
                            if result_data == None:
                                log.error("CloudWorker starting failed ! Unassign it and skipping this batch")
                                db.sql_update("CloudWorker",('batch_id',),("NULL",),"cloud_worker_id='%s'" % cw['cloud_worker_id'])
                            else:
                                if result_data != True:
                                    log.info("Updating its data to %s" % (result_data))
                                    db.sql_update("CloudWorker",('cloud_worker_data',),(result_data,),"cloud_worker_id='%s'" % cw['cloud_worker_id'])

                                #TODO is this a bad hack ?
                                if len(cworkers)==0:
                                    log.info("Fist Cloudworker started! Configuring the grid to enable QoS for batch %s" % (b['batch_id']))
                                    dg_rpc.configure_qos(b['batch_id'],b['dg_id'],b['dg_conf_file'],cw['cloud_worker_id'])
                                                                    
                                #TODO ? First billing of the cloud worker use
            log.info("End of managment for batch %s" % b['batch_id'])



if __name__=="__main__":
    #To use a cloud worker, the cloud assisted makespan should be MIN_BENEFIT times lower than normal makespan (with tail problem)
    MIN_BENEFIT=1.5

    db = db_func.SpeQuloS_DB(os.path.join(sys.path[0],"db.cfg"))        
    monitor_batches(db)



